# Copyright Splunk Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import logging
from functools import partial
from os import environ
from typing import Callable, Collection, Dict, List, Optional, Tuple, Union

from opentelemetry.environment_variables import OTEL_TRACES_EXPORTER
from opentelemetry.instrumentation.propagators import ResponsePropagator
from opentelemetry.instrumentation.version import (
    __version__ as auto_instrumentation_version,
)
from opentelemetry.sdk.environment_variables import (
    OTEL_ATTRIBUTE_COUNT_LIMIT,
    OTEL_ATTRIBUTE_VALUE_LENGTH_LIMIT,
    OTEL_EVENT_ATTRIBUTE_COUNT_LIMIT,
    OTEL_LINK_ATTRIBUTE_COUNT_LIMIT,
    OTEL_SPAN_ATTRIBUTE_COUNT_LIMIT,
    OTEL_SPAN_EVENT_COUNT_LIMIT,
    OTEL_SPAN_LINK_COUNT_LIMIT,
)
from opentelemetry.sdk.resources import Resource
from opentelemetry.sdk.trace.export import SpanExporter
from pkg_resources import iter_entry_points

from splunk_otel.environment_variables import (
    _SPLUNK_ACCESS_TOKEN,
    _SPLUNK_TRACE_RESPONSE_HEADER_ENABLED,
)
from splunk_otel.propagators import _ServerTimingResponsePropagator
from splunk_otel.symbols import (
    _DEFAULT_EXPORTERS,
    _DEFAULT_MAX_ATTR_LENGTH,
    _DEFAULT_OTEL_SERVICE_NAME,
    _DEFAULT_SERVICE_NAME,
    _DEFAULT_SPAN_LINK_COUNT_LIMIT,
    _EXPORTER_OTLP,
    _EXPORTER_OTLP_GRPC,
    _KNOWN_EXPORTER_PACKAGES,
    _LIMIT_UNSET_VALUE,
    _NO_SERVICE_NAME_WARNING,
    _SERVICE_NAME_ATTR,
    _SPLUNK_DISTRO_VERSION_ATTR,
    _TELEMETRY_VERSION_ATTR,
)
from splunk_otel.version import __version__

_SpanExporterFactory = Callable[["_Options"], SpanExporter]
_SpanExporterClass = Callable[..., SpanExporter]

logger = logging.getLogger("options")


class _Options:
    span_exporter_factories: Collection[_SpanExporterFactory]
    access_token: Optional[str]
    resource: Resource
    response_propagation: bool
    response_propagator: Optional[ResponsePropagator]

    def __init__(
        self,
        service_name: Optional[str] = None,
        span_exporter_factories: Optional[Collection[_SpanExporterFactory]] = None,
        access_token: Optional[str] = None,
        resource_attributes: Optional[Dict[str, Union[str, bool, int, float]]] = None,
        trace_response_header_enabled: Optional[bool] = None,
    ):
        self._set_default_env()
        self.access_token = self._get_access_token(access_token)
        self.response_propagator = self._get_trace_response_header_enabled(
            trace_response_header_enabled
        )
        self.resource = self._get_resource(service_name, resource_attributes)
        self.span_exporter_factories = self._get_span_exporter_factories(
            span_exporter_factories
        )

    @staticmethod
    def _get_access_token(access_token: Optional[str]) -> Optional[str]:
        if not access_token:
            access_token = environ.get(_SPLUNK_ACCESS_TOKEN)
        return access_token or None

    @staticmethod
    def _get_trace_response_header_enabled(
        enabled: Optional[bool],
    ) -> Optional[ResponsePropagator]:
        if enabled is None:
            enabled = _Options._is_truthy(
                environ.get(_SPLUNK_TRACE_RESPONSE_HEADER_ENABLED, "true")
            )
        if enabled:
            return _ServerTimingResponsePropagator()
        return None

    @staticmethod
    def _get_resource(
        service_name: Optional[str],
        attributes: Optional[Dict[str, Union[str, bool, int, float]]],
    ) -> Resource:
        attributes = attributes or {}
        if service_name:
            attributes[_SERVICE_NAME_ATTR] = service_name
        attributes.update(
            {
                _TELEMETRY_VERSION_ATTR: auto_instrumentation_version,
                _SPLUNK_DISTRO_VERSION_ATTR: __version__,
            }
        )
        resource = Resource.create(attributes)
        if (
            resource.attributes.get(_SERVICE_NAME_ATTR, _DEFAULT_OTEL_SERVICE_NAME)
            == _DEFAULT_OTEL_SERVICE_NAME
        ):
            logger.warning(_NO_SERVICE_NAME_WARNING)
            resource = resource.merge(
                Resource({_SERVICE_NAME_ATTR: _DEFAULT_SERVICE_NAME})
            )
        return resource

    @staticmethod
    def _get_span_exporter_factories(
        factories: Optional[Collection[_SpanExporterFactory]],
    ) -> Collection[_SpanExporterFactory]:
        if factories:
            return factories

        exporter_names = _Options._get_span_exporter_names_from_env()
        return _Options._import_span_exporter_factories(exporter_names)

    @classmethod
    def _is_truthy(cls, value: Optional[str]) -> bool:
        if not value:
            return False

        return value.strip().lower() not in (
            "false",
            "no",
            "f",
            "0",
        )

    @staticmethod
    def _set_default_env() -> None:
        defaults = {
            OTEL_ATTRIBUTE_COUNT_LIMIT: _LIMIT_UNSET_VALUE,
            OTEL_SPAN_ATTRIBUTE_COUNT_LIMIT: _LIMIT_UNSET_VALUE,
            OTEL_SPAN_EVENT_COUNT_LIMIT: _LIMIT_UNSET_VALUE,
            OTEL_EVENT_ATTRIBUTE_COUNT_LIMIT: _LIMIT_UNSET_VALUE,
            OTEL_LINK_ATTRIBUTE_COUNT_LIMIT: _LIMIT_UNSET_VALUE,
            OTEL_SPAN_LINK_COUNT_LIMIT: str(_DEFAULT_SPAN_LINK_COUNT_LIMIT),
            OTEL_ATTRIBUTE_VALUE_LENGTH_LIMIT: str(_DEFAULT_MAX_ATTR_LENGTH),
        }

        for key, value in defaults.items():
            if key not in environ:
                environ[key] = value

    @classmethod
    def _get_span_exporter_names_from_env(cls) -> Collection[Tuple[str, str]]:
        exporters_env = (
            environ.get(OTEL_TRACES_EXPORTER, "").strip() or _DEFAULT_EXPORTERS
        )

        exporters: List[Tuple[str, str]] = []
        if not cls._is_truthy(exporters_env):
            return exporters

        # exporters are known by different names internally by Python Otel SDK.
        # Here we create a mapping of user provided names to internal names so
        # that we can provide helpful error messages later.
        for name in exporters_env.split(","):
            if name == _EXPORTER_OTLP:
                exporters.append((_EXPORTER_OTLP, _EXPORTER_OTLP_GRPC))
            else:
                exporters.append(
                    (
                        name,
                        name,
                    )
                )
        return exporters

    @staticmethod
    def _import_span_exporter_factories(
        exporter_names: Collection[Tuple[str, str]],
    ) -> Collection[_SpanExporterFactory]:
        factories = []
        entry_points = {
            ep.name: ep for ep in iter_entry_points("opentelemetry_traces_exporter")
        }

        for name, internal_name in exporter_names:
            if internal_name not in entry_points:
                package = _KNOWN_EXPORTER_PACKAGES.get(internal_name)
                if package:
                    help_msg = f"please make sure {package} is installed"
                else:
                    help_msg = (
                        "please make sure the relevant exporter package is installed."
                    )
                raise ValueError(
                    f'exporter "{name} ({internal_name})" not found. {help_msg}'
                )

            exporter_class: _SpanExporterClass = entry_points[internal_name].load()
            if internal_name == _EXPORTER_OTLP_GRPC:
                factory = _Options._otlp_factory
            else:
                factory = _Options._generic_exporter
            factories.append(partial(factory, exporter_class))
        return factories

    @staticmethod
    def _generic_exporter(
        exporter: _SpanExporterClass,
        options: "_Options",  # pylint: disable=unused-argument
    ) -> SpanExporter:
        return exporter()

    @staticmethod
    def _otlp_factory(exporter: _SpanExporterClass, options: "_Options") -> SpanExporter:
        return exporter(headers=(("x-sf-token", options.access_token),))



# Copyright Splunk Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import typing

from opentelemetry import trace
from opentelemetry.context.context import Context
from opentelemetry.instrumentation.propagators import (
    _HTTP_HEADER_ACCESS_CONTROL_EXPOSE_HEADERS,
    ResponsePropagator,
    default_setter,
)
from opentelemetry.propagators import textmap
from opentelemetry.trace import format_span_id, format_trace_id


class _ServerTimingResponsePropagator(ResponsePropagator):
    """Response propagator that injects tracecontext into HTTP responses as Server-Timing headers."""

    def inject(
        self,
        carrier: textmap.CarrierT,
        context: typing.Optional[Context] = None,
        setter: textmap.Setter = default_setter,  # type: ignore
    ) -> None:
        """Injects SpanContext into the HTTP response carrier."""
        span = trace.get_current_span(context)
        span_context = span.get_span_context()
        if span_context == trace.INVALID_SPAN_CONTEXT:
            return

        header_name = "Server-Timing"
        flags = span_context.trace_flags
        trace_id = format_trace_id(span_context.trace_id)
        span_id = format_span_id(span_context.span_id)

        setter.set(
            carrier,
            header_name,
            f'traceparent;desc="00-{trace_id}-{span_id}-{flags:02x}"',
        )
        setter.set(
            carrier,
            _HTTP_HEADER_ACCESS_CONTROL_EXPOSE_HEADERS,
            header_name,
        )



# Copyright Splunk Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import logging
import os
from typing import Any


def _is_truthy(value: Any) -> bool:
    if isinstance(value, str):
        value = value.lower().strip()
    return value in [True, 1, "true", "yes"]


def _get_log_level(level):
    levels = {
        "none": logging.NOTSET,
        "debug": logging.DEBUG,
        "info": logging.INFO,
        "warn": logging.WARNING,
        "error": logging.ERROR,
        "fatal": logging.CRITICAL,
    }

    return levels[level.lower()]


def _init_logger(name):
    level = _get_log_level(os.environ.get("OTEL_LOG_LEVEL", "info"))
    logger = logging.getLogger(name)
    logger.setLevel(level)
    logger.addHandler(logging.StreamHandler())



# Copyright Splunk Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import logging
import os

from opentelemetry.sdk.metrics import MeterProvider

from splunk_otel.util import _is_truthy

logger = logging.getLogger(__name__)


def start_metrics() -> MeterProvider:
    # pylint: disable=import-outside-toplevel
    from opentelemetry.instrumentation.system_metrics import SystemMetricsInstrumentor

    enabled = os.environ.get("OTEL_METRICS_ENABLED", True)
    if not _is_truthy(enabled):
        logger.info("metering has been disabled with OTEL_METRICS_ENABLED=%s", enabled)
        return None

    try:
        meter_provider = _configure_metrics()
        system_metrics = SystemMetricsInstrumentor()
        system_metrics.instrument(meter_provider=meter_provider)
        logger.debug("Instrumented runtime metrics")
        return meter_provider
    except Exception as exc:  # pylint:disable=broad-except
        logger.exception("Instrumenting of runtime metrics failed")
        raise exc


def _configure_metrics() -> MeterProvider:
    # pylint: disable=import-outside-toplevel
    from opentelemetry.exporter.otlp.proto.grpc.metric_exporter import (
        OTLPMetricExporter,
    )
    from opentelemetry.metrics import set_meter_provider
    from opentelemetry.sdk.metrics.export import PeriodicExportingMetricReader

    metrics_exporter = OTLPMetricExporter()
    meter_provider = MeterProvider([PeriodicExportingMetricReader(metrics_exporter)])
    set_meter_provider(meter_provider)
    return meter_provider
